//
//  CLNTextProcessor.m
//  Collocations
//
//  Created by Abraham Avnisan on 11/10/15.
//  Copyright (c) 2015 Abraham Avnisan. All rights reserved.
//

#import "CLNTextProcessor.h"
#import "CLNSubtextWordSegment.h"
#import "CLNSubtextWord.h"
#import "enumsAndConstants.h"
#import "HelperFunctions.h"

@interface CLNTextProcessor ()

@property (strong, nonatomic) NSRegularExpression *RE_SpecialChars;
@property (strong, nonatomic) NSRegularExpression *RE_DoubleAsterisk;
@property (strong, nonatomic) NSRegularExpression *RE_ContentsOfDoubleAsterisks;
@property (strong, nonatomic) NSRegularExpression *RE_ContentsOfDoubleAsterisksInclusive;
@property (strong, nonatomic) NSRegularExpression *RE_ContentsOfBrackets;
@property (strong, nonatomic) NSRegularExpression *RE_PreviousWordFlag;
@property (strong, nonatomic) NSRegularExpression *RE_ProbabilityCurvature;

@end

@implementation CLNTextProcessor

#pragma mark - main methods

- (CLNPage* )processRawTextFile:(NSString *)fileName forTextBelongingToBohr:(BOOL)bohrIsAuthor
{
    
    // declare local vars
    CLNPage* page = nil;
    
    NSString *filePath = [[NSBundle mainBundle] pathForResource:fileName ofType:@"txt"];
    
    // read everything from text
    NSString* fileContents = [NSString stringWithContentsOfFile:filePath
                                                       encoding:NSUTF8StringEncoding
                                                          error:nil];
    
    // if file contents were extracted, create a pageObj
    if (fileContents) {
        page = [[CLNPage alloc] initWithFileContents:fileContents author:bohrIsAuthor];
    }
    
    // populate the page model with subtext info from markdown files
    [self processMarkdownRootForPage:page];
    
    return page;
}

- (void)processMarkdownRootForPage:(CLNPage *)page
{
    // declare local vars
    BOOL bohrIsAuthor = page.bohrIsAuthor;
    
    // open root MD folder and store its contents in an array
    NSString *bundlePath = [[NSBundle mainBundle] bundlePath];
    NSString *MDpath = [bundlePath stringByAppendingString:@ROOT_MD_DIR];
    
    NSError *error;
    NSArray *dirContents = [[NSFileManager defaultManager] contentsOfDirectoryAtPath:MDpath error:&error];
    
    if (error) NSLog(@"error: %@", [error localizedDescription]);
    
    // iterate over the dirContents array
    for (NSString *fileName in dirContents) {
        
        // grab basic info about this file from its prefix
        struct MDPrefix prefix = [self extractMDFilePrefixFrom:fileName];
        
        // only process MD files for the current author
        if (prefix.bohrIsAuthor == bohrIsAuthor) {
            
            // create file path
            NSString *path = [NSString stringWithFormat:@"%@/%@", MDpath, fileName];
            
            
            NSString *markdown = [NSString stringWithContentsOfFile:path
                                                           encoding:NSUTF8StringEncoding
                                                              error:NULL];
            
            // strip markdown of html tags
            NSString *markdownClean = [self stringByStrippingHTMLFrom:markdown];
            
            
            // OK, let's process that markdown, bitches!
            [self processMD:markdownClean withPrefix:prefix forPage:page];
            
        }
    }
}

- (void)processMD:(NSString *)markdown withPrefix:(struct MDPrefix)prefix forPage:(CLNPage *)page
{
    
    // make sure markdown contains the *exact* same number of
    // words, lines, and chars as our page object
    [self verifyMD:markdown forPage:page];
    
    // grab a reference to the appropriate subtext array from the pages's mutli-dimensional subtext array
    // (we are going to create CLNSubtextWord objects based on the MD and add them to this array)
    NSMutableArray *subtextWordArray = page.subtextsAs2DArray[prefix.xPosition][prefix.yPosition];
    
    NSMutableArray *subtextWordSegmentArray = [[NSMutableArray alloc] init];
    
    // create mutable version of markdown without new line characters
    NSMutableString *markdownMutable = [[NSMutableString alloc] initWithString:[markdown stringByReplacingOccurrencesOfString:@"\n"
                                                                                                                   withString:@" "]];

    BOOL firstTimeInLoop = true;
    
    // iterate over all the subtexts as defined by starting and ending double asterisk pair
    while ([self.RE_ContentsOfDoubleAsterisksInclusive numberOfMatchesInString:markdownMutable
                                                                       options:0
                                                                         range:hfGetFullRangeOfString(markdownMutable)] > 0) {
        
        // init local vars
        NSTextCheckingResult *result = [self.RE_ContentsOfDoubleAsterisksInclusive firstMatchInString:markdownMutable
                                                                                              options:0
                                                                                                range:hfGetFullRangeOfString(markdownMutable)];
        
        NSMutableString *matchMutable = [[NSMutableString alloc] initWithString:[markdownMutable substringWithRange:result.range]];
        
        // check if match is connected to previous segment
        NSUInteger matches = [self.RE_PreviousWordFlag numberOfMatchesInString:matchMutable
                                                                       options:0
                                                                         range:hfGetFullRangeOfString(matchMutable)];
        
        BOOL linkedToPrevious = matches > 0 ? true : false;
        
        // if this match isn't linked to previous (and this isn't the first loop)
        // we can finish processing segments stored from previous iterations of
        // the loop before moving on
        if (!linkedToPrevious && !firstTimeInLoop > 0) {
            
            // combine segments into subtext
            CLNSubtextWord* subtextWord = [self returnSubtextWordByCombiningSegmentsInArray:subtextWordSegmentArray forPage:page];
            
            // add to the page's subtext word array
            [subtextWordArray addObject:subtextWord];
            
            // reset our subtextWordSegmentArray
            [subtextWordSegmentArray removeAllObjects];
            
        }
        
        
        
        // SET PROBABILITY CURVATURE
        // _________________________________ //
        
        NSNumber *probCurvature = [self extractProbCurvatureFrom:matchMutable];
        
        // if prob curvature is not found in the string, use previous segment or set a default
        if (!probCurvature) {
            if (linkedToPrevious) {
                probCurvature = [(CLNSubtextWordSegment *)[subtextWordSegmentArray lastObject] probabilityCurvature];
            } else {
                probCurvature = [[NSNumber alloc] initWithFloat:DEFAULT_PROB_CURVATURE];
            }
        }
        
        // CALCULATE RANGE IN ORIGINAL
        // _________________________________ //
        
        // get length of string without special chars

        // strip special chars in this segment from markdown
        [self.RE_SpecialChars replaceMatchesInString:markdownMutable
                                             options:0
                                               range:result.range
                                        withTemplate:@""];
        
        // strip special chars from matchMutable
        [self.RE_SpecialChars replaceMatchesInString:matchMutable
                                             options:0
                                               range:hfGetFullRangeOfString(matchMutable)
                                        withTemplate:@""];
        
        NSUInteger strippedMatchLength = [matchMutable length];
        
        NSRange rangeInOriginal = NSMakeRange(result.range.location, strippedMatchLength);
        
        NSString *segmentString = [markdownMutable substringWithRange:rangeInOriginal];

        // create a CLNSubtextWord object and add it to our array
        CLNSubtextWordSegment* segment = [[CLNSubtextWordSegment alloc] initWithRangeInOriginal:rangeInOriginal
                                                                           probabilityCurvature:probCurvature
                                                                                   linkedToPrev:linkedToPrevious
                                                                                      forString:segmentString];

        
        [subtextWordSegmentArray addObject:segment];
        
        firstTimeInLoop = false;
        
    }
    
    // there will always be one more subtext to create after this loop...

    // combine segments into subtext
    CLNSubtextWord* subtextWord = [self returnSubtextWordByCombiningSegmentsInArray:subtextWordSegmentArray forPage:page];
    
    // add to the page's subtext word array
    [subtextWordArray addObject:subtextWord];
    
    [self logSubtextWordArray:subtextWordArray forPage:page];
    
}

#pragma mark - helper methods

- (CLNSubtextWord *)returnSubtextWordByCombiningSegmentsInArray:(NSArray *)segments forPage:(CLNPage *)page
{
    
    // declare vars we need to init our CLNSubtextWord object
    // __________________________________________________________________ //
    
    NSString       *subtextAsString = @"";
    NSMutableArray *probCurvatures  = [[NSMutableArray alloc] init];
    NSMutableArray *charIDs         = [[NSMutableArray alloc] init];
    NSMutableArray *lineIndices     = [[NSMutableArray alloc] init];
    NSMutableArray *wordIndices     = [[NSMutableArray alloc] init];
    
    NSPointerArray *pointersToLines = [NSPointerArray weakObjectsPointerArray];
    NSPointerArray *pointersToWords = [NSPointerArray weakObjectsPointerArray];
    NSPointerArray *pointersToChars = [NSPointerArray weakObjectsPointerArray];
    
    // iterate over segments
    // __________________________________________________________________ //
    
    for (CLNSubtextWordSegment* segment in segments) {
        
        // SET SUBTEXT
        subtextAsString = [subtextAsString stringByAppendingString:segment.segmentAsString];
        
        // SET PROB CURVATURE
        [probCurvatures addObject:segment.probabilityCurvature];
        
        // SET CHARIDs
        // iterate over chars in this segment and add them to charIDs array
        NSUInteger indexEnd = segment.rangeInOriginal.location + segment.rangeInOriginal.length;
        for (NSUInteger i = segment.rangeInOriginal.location; i < indexEnd; i++) {
            NSNumber* charID = [NSNumber numberWithLongLong:i];
            [charIDs addObject:charID];
        }
        
        // figure out which line(s) & word(s) this segment includes
        // __________________________________________________________________ //
        
        int firstCharIDSegment = segment.rangeInOriginal.location;
        int lastCharIDSegment =  segment.rangeInOriginal.location + segment.rangeInOriginal.length - 1;
        
        // iterate over every line in the page until this segment is located
        for (int i = 0; i < [page.lineObjects count]; i++) {
            
            BOOL finishedWithThisSegment = false;
            
            // check if this segment falls within this line
            CLNLine* lineObj = page.lineObjects[i];
            
            CLNWord* firstWord = [lineObj.wordObjects firstObject];
            CLNWord* lastWord = [lineObj.wordObjects lastObject];
            
            CLNChar* firstChar = [firstWord.charObjects firstObject];
            CLNChar* lastChar = [lastWord.charObjects lastObject];
            
            int firstCharIDLine = [firstChar.uniqueID intValue];
            int lastCharIDLine = [lastChar.uniqueID intValue];
            
            
            // check if segment **begins** in this line
            
            BOOL segmentBeginsInThisLine = false;
            BOOL segmentEndsInThisLine = false;
            
            if (firstCharIDLine <= firstCharIDSegment && firstCharIDSegment <= lastCharIDLine) {
                segmentBeginsInThisLine = true;
            }
            
            if (lastCharIDSegment >= firstCharIDLine && lastCharIDSegment <= lastCharIDLine) {
                segmentEndsInThisLine = true;
            }
            
            if (segmentBeginsInThisLine || segmentEndsInThisLine) {
                
                // add this line index to our array
                BOOL addToLineIndexArray = true;
                NSNumber *lineIndex = [NSNumber numberWithInt:i];
                
                // don't add the index to the array if it already exists
                if ([lineIndices count] > 0) {
                    
                    NSNumber *previousLineIndex = (NSNumber *)[lineIndices lastObject];
                    if ([lineIndex intValue] == [previousLineIndex intValue]) {
                        addToLineIndexArray = false;
                    }
                }
                
                if (addToLineIndexArray) {
                    [lineIndices addObject:lineIndex];
                    [pointersToLines addPointer:(__bridge void *)lineObj];
                }
                
                // iterate over words in the line
                for (int j = 0; j  < [lineObj.wordObjects count]; j++) {
                    
                    NSString* thisWordAsString = [(CLNWord *)lineObj.wordObjects[j] wordAsString];
                    
                    BOOL thisWordIndexAddedToArray = false;
                    
                    // check if this segment falls within this word
                    CLNWord* wordObj = (CLNWord *)lineObj.wordObjects[j];
                    
                    int firstCharIDWord = [(NSNumber *)[(CLNChar *)[wordObj.charObjects firstObject] uniqueID] intValue];
                    int lastCharIDWord  = [(NSNumber *)[(CLNChar *)[wordObj.charObjects lastObject] uniqueID] intValue];
                    
                    // check if segment *begins* in this word
                    if (firstCharIDWord <= firstCharIDSegment && firstCharIDSegment <= lastCharIDWord) {
                        
                        // add word index to our array
                        BOOL addToWordIndexArray = true;
                        
                        NSNumber *wordIndex = wordObj.wordID;
                        
                        // don't add the index to the array if it already exists
                        if ([wordIndices count] > 0) {
                            
                            NSNumber *previousWordIndex = (NSNumber *)[wordIndices lastObject];
                            if ([wordIndex intValue] == [previousWordIndex intValue]) {
                                addToWordIndexArray = false;
                            }
                        }
                        
                        if (addToWordIndexArray) {
                            [wordIndices addObject:wordIndex];
                            [pointersToWords addPointer:(__bridge void *)wordObj];
                            
                        }
                        thisWordIndexAddedToArray = true;
                        
                    } // END check if segment *begins* in this word
                    
                    // if word has not been added to array,
                    // check if segment *ends* in this word
                    if (lastCharIDSegment <= firstCharIDWord && lastCharIDSegment <= lastCharIDWord) {
                        
                        if (!thisWordIndexAddedToArray) {
                            
                            NSNumber *wordIndex = wordObj.wordID;
                            [wordIndices addObject:wordIndex];
                            thisWordIndexAddedToArray = true;
                        }
                    }
                    
                    if (thisWordIndexAddedToArray) {
                        finishedWithThisSegment = true;
                        break;
                    }
                } // END iterate over every word
                
                if (finishedWithThisSegment) {
                    break;
                }
                
                // if segment *ends* in this line break out of the for loop
                if (lastCharIDSegment <= lastCharIDLine) break;
                
            } // END if segment ends or begins in line
        } // END iterate over every line
    } // END iterate over segements
    
    
    // check to see which CLNChar objects
    // should be associated with this subtext
    // __________________________________________________________________ //
    
    for (CLNWord *wordObj in pointersToWords) {
        
        for (CLNChar* charObj in wordObj.charObjects) {
            
            NSNumber* charID = charObj.uniqueID;
            if ([charIDs containsObject:charID]) {
                charObj.isActive = true;
            }
            
        }
        
    }
    
    // MAKE AND RETURN OUR SUBTEXT
    // __________________________________________________________________ //
    
    // grab non-default prob curvature if there is one
    float probCurvature = [self getCorrectProbCurvatureFromArray:probCurvatures];
    
    // init our subtext
    CLNSubtextWord *subtext = [[CLNSubtextWord alloc] initWithUniqueCharIDs:charIDs
                                                            pointersToLines:pointersToLines
                                                                wordIndices:wordIndices
                                                            pointersToWords:pointersToWords
                                                                lineIndices:lineIndices
                                                                  forString:subtextAsString
                                                              probCurvature:probCurvature];
    
    // and... return it
    return subtext;
}
- (float)getCorrectProbCurvatureFromArray:(NSArray *)array
{
    float probCurvature = DEFAULT_PROB_CURVATURE;
    
    for (NSNumber* curvature in array) {
        if ([curvature floatValue] != probCurvature) {
            probCurvature = [curvature floatValue];
        }
    }
    
    return probCurvature;
}

- (NSNumber *)extractProbCurvatureFrom:(NSString *)string
{
    NSNumber *probabilityCurvature = nil;
    NSRange nullRange = NSMakeRange(0, 0);
    
    NSTextCheckingResult *probability = [self.RE_ProbabilityCurvature firstMatchInString:string
                                                                                 options:0
                                                                                   range:hfGetFullRangeOfString(string)];
    
    if ((probability.range.location != nullRange.location) && (probability.range.length != nullRange.length)) {
        
        NSString *probabilityAsStr = [string substringWithRange:probability.range];
        hfLogNSRange(probability.range);
        probabilityCurvature = [[NSNumber alloc] initWithFloat:[probabilityAsStr floatValue]];
    }
    
    return probabilityCurvature;
    
}

- (void)verifyMD:(NSString *)markdown forPage:(CLNPage *)page
{
    // GET PAGE INFO
    // _________________________________ //
    
    // get number of lines from page
    int numLinesPage = (int)[page.lineObjects count];
    
    // calculate number of words and characters in page
    int numWordsPage = 0;
    int numCharsPage = 0;

    for (CLNLine* lineObj in page.lineObjects) {
        
        numWordsPage += [lineObj.wordObjects count];
        
        for (CLNWord* wordObj in lineObj.wordObjects) {
            
            numCharsPage += wordObj.numChars;
        }
    }
    
    
    // GET MARKDOWN INFO
    // _________________________________ //
    
    // grab a cleaned up version of the markdown
    NSString *markdownClean = [self stringByStrippingSpecialCharsFrom:markdown];
    
    // get number of lines in markdown
    NSArray *lines = [markdownClean componentsSeparatedByString:@"\n"];
    
    // only count non-emtpy lines
    int numLinesMD = (int)[lines count];

    // get number of words in md
    NSString *mdLinesStripped = [markdownClean stringByReplacingOccurrencesOfString:@"\n" withString:@" "];
    
    int numWordsMD = (int)[[mdLinesStripped componentsSeparatedByString:@" "] count];
    
    // get number of chars in md
    NSString *mdWordsSpacesStripped = [mdLinesStripped stringByReplacingOccurrencesOfString:@" " withString:@""];
    int numCharsMD = (int)[mdWordsSpacesStripped length];
    
    
    // & NOW....COMPARE THE TWO
    // _________________________________ //
    
    if (numLinesPage != numLinesMD || numWordsPage != numWordsMD || numCharsPage != numCharsMD) {
        NSLog(@"ERROR! Markdown file's lines, words, or chars count to not match page's lines, words or char count!!");
        exit(EXIT_FAILURE);
    }
    
    NSLog(@"SUCCESS - Markdown file line, word and char count verified!");

}
- (struct MDPrefix)extractMDFilePrefixFrom:(NSString *)file
{
    struct MDPrefix prefix;
    
    NSRegularExpression *RE_filePrefix = [[NSRegularExpression alloc] initWithPattern:@"(?<=\\().+(?=\\))"
                                                                              options:NSRegularExpressionCaseInsensitive
                                                                                error:NULL];
    
    
    
    NSTextCheckingResult* result = [RE_filePrefix firstMatchInString:file
                                                             options:0
                                                               range:NSMakeRange(0, [file length])];
    
    NSString *prefixAsStr = [file substringWithRange:result.range];
    
    // grab components from prefix
    NSArray *prefixComponents = [prefixAsStr componentsSeparatedByString:@"_"];
    
    /*  we require files to be saved following these rules
     
     PROTOCOL: (authorFirstLetter_xPosition_yPosition) optional description.md
     EXAMPLE: (b_3_7) i tried each thing.md  */
    
    // if there aren't 3 prefix components throw an error
    if ([prefixComponents count] != 3) {
        NSLog(@"ERROR: Markdown file not saved correctly");
    }
    
    // store author
    if ([(NSString *)prefixComponents[0] isEqualToString:@"b"]) {
        prefix.bohrIsAuthor = true;
    } else if ([(NSString *)prefixComponents[0] isEqualToString:@"e"]) {
        prefix.bohrIsAuthor = false;
    }
    
    // store xPosition
    prefix.xPosition = [(NSString *)prefixComponents[1] intValue];
    
    // store yPositoin
    prefix.yPosition = [(NSString *)prefixComponents[2] intValue];
    
    return prefix;
    
}
- (NSString *)stringByStrippingSpecialCharsFrom:(NSString *)string
{
    
    NSMutableString *stringMutable = [[NSMutableString alloc] initWithString:string];
    
    [self.RE_SpecialChars replaceMatchesInString:stringMutable
                                         options:0
                                           range:NSMakeRange(0, [stringMutable length])
                                    withTemplate:@""];
    
    // remove spaces and new line characters from beginning and end of string
    NSString *returnString = [stringMutable stringByTrimmingCharactersInSet:[NSCharacterSet whitespaceAndNewlineCharacterSet]];
    
    return returnString;
    
}
- (NSString *)stringByStrippingHTMLFrom:(NSString *)string
{
    NSMutableString *stringMutable = [[NSMutableString alloc] initWithString:string];
    
    NSRegularExpression *RE_HTMLTags = [NSRegularExpression regularExpressionWithPattern:@"<[^-].+?>"
                                                                                options:0
                                                                                  error:nil];
   [RE_HTMLTags replaceMatchesInString:stringMutable
                               options:0
                                 range:NSMakeRange(0, [stringMutable length])
                          withTemplate:@""];
    
    NSString* returnString = [[NSString alloc] initWithString:stringMutable];
    
    return returnString;
    
}

#pragma mark - setters & getters

- (NSRegularExpression *)RE_SpecialChars
{
    if (!_RE_SpecialChars) {
        _RE_SpecialChars = [NSRegularExpression regularExpressionWithPattern:@"(\\*)|(\\[.+?])"
                                                                     options:0
                                                                       error:nil];
    }
    return _RE_SpecialChars;
}
- (NSRegularExpression *)RE_DoubleAsterisk
{
    if (!_RE_DoubleAsterisk) {
        _RE_DoubleAsterisk = [NSRegularExpression regularExpressionWithPattern:@"\\*\\*"
                                                                       options:0
                                                                         error:nil];
    }
    return _RE_DoubleAsterisk;
}
- (NSRegularExpression *)RE_ContentsOfDoubleAsterisks
{
    if (!_RE_ContentsOfDoubleAsterisks) {
        _RE_ContentsOfDoubleAsterisks = [NSRegularExpression regularExpressionWithPattern:@"(?<=\\*\\*).+?(?=\\*\\*)"
                                                                                  options:0
                                                                                    error:nil];
    }
    return _RE_ContentsOfDoubleAsterisks;
}
- (NSRegularExpression *)RE_ContentsOfDoubleAsterisksInclusive
{
    if (!_RE_ContentsOfDoubleAsterisksInclusive) {
        _RE_ContentsOfDoubleAsterisksInclusive = [NSRegularExpression regularExpressionWithPattern:@"\\*\\*.+?\\*\\*"
                                                                                           options:0
                                                                                             error:nil];
    }
    return _RE_ContentsOfDoubleAsterisksInclusive;
}
- (NSRegularExpression *)RE_ContentsOfBrackets
{
    if (!_RE_ContentsOfBrackets) {
        _RE_ContentsOfBrackets = [NSRegularExpression regularExpressionWithPattern:@"(?<=\\[).+?(?=])"
                                                                           options:0
                                                                             error:nil];
    }
    return _RE_ContentsOfBrackets;
}
- (NSRegularExpression *)RE_PreviousWordFlag
{
    if (!_RE_PreviousWordFlag) {
        _RE_PreviousWordFlag = [NSRegularExpression regularExpressionWithPattern:@"<-"
                                                                         options:0
                                                                           error:nil];
    }
    return _RE_PreviousWordFlag;
}
- (NSRegularExpression *)RE_ProbabilityCurvature
{
    if (!_RE_ProbabilityCurvature) {
        _RE_ProbabilityCurvature = [NSRegularExpression regularExpressionWithPattern:@"\\.[0-9]+"
                                                                             options:0
                                                                               error:nil];
    }
    return _RE_ProbabilityCurvature;
}

@end
